/* required for popen on some systems */
#define _XOPEN_SOURCE

#ifdef WIN32
# if defined(__MINGW32__)
#  include <sys/stat.h>
# endif
# include <windows.h>
#else
# include <sys/stat.h>
# include <sys/time.h>
# include <unistd.h>
#endif

#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <time.h>
#include <errno.h>

#ifndef WIN32
#include <sys/wait.h>
#endif

#include <dirent.h>

#include "stutter.h"
#include "compat.h"

#if defined(__CYGWIN__)
# define timezone _timezone
#endif

/* get value of next parameter (returns 1 on success, 0 if there are no more) */
STT_INLINE int stt_nextparm(stt_obj_t **obj, stt_obj_t **parm)
{
	if((*parm)->type != STT_ST_CONS) return 0;
	*obj = (*parm)->d.cons.car;
	*parm = (*parm)->d.cons.cdr;
	return 1;
}

/* get and evaluate next parameter (returns 1 on success, 0 if there are no more) */
STT_INLINE int stt_nextarg(stt_varctx_t *ctx, stt_obj_t **obj, stt_obj_t **parm)
{
	if(!stt_nextparm(obj, parm)) return 0;
	*obj = stt_eval(ctx, *obj);
	return 1;
}

/* get and evaluate next parameter (returns 1 on success, 0 if there are no more).
 * if it's the last parameter and tail is true, calls eval
 * with tail flag set to 1! */
STT_INLINE int stt_nextargtail(stt_varctx_t *ctx, stt_obj_t **obj, stt_obj_t **parm, const int tail)
{
	if(!stt_nextparm(obj, parm)) return 0;
	*obj = stt_eval_(ctx, *obj, tail && (!(*parm) || (*parm)->type != STT_ST_CONS));
	return 1;
}

/* get and evaluate next parameter, ensure it's from a type
 * returns 1 on success */
int stt_next_t(stt_varctx_t *ctx, stt_obj_t **obj, stt_obj_t **parm, const int type)
{
	if(!stt_nextarg(ctx, obj, parm)) return 0;
	if((*obj)->type != type) return 0;
	return 1;
}


#include "builtin/basic_impl.c"
#include "builtin/list_impl.c"
#include "builtin/type_impl.c"
#include "builtin/algebra_impl.c"
#include "builtin/interp_impl.c"
#include "builtin/str_impl.c"
#include "builtin/io_impl.c"
#include "builtin/array_impl.c"
#include "builtin/dict_impl.c"
#include "builtin/date_impl.c"
#include "builtin/clos_impl.c"
#include "builtin/os_impl.c"
#include "builtin/popen_impl.c"
#include "builtin/blob_impl.c"
#include "builtin/sforms_impl.c"
#include "builtin/math_impl.c"


/* register builtins in variable context */
void stt_register_builtins(stt_varctx_t *ctx)
{
	stt_s_mark mark;

	mark = stt_gc_prot_mark(ctx->stt);

	/* constants */
	stt_varctx_set(ctx, "NIL", stt_create_nil(ctx->stt));
	stt_varctx_set(ctx, "T", stt_create_t(ctx->stt));
	stt_varctx_set(ctx, "EOF", stt_create_eof(ctx->stt));

	stt_varctx_set(ctx, "NEW-NIL", stt_create_builtin(ctx->stt, (stt_builtin_handler) stt_create_new_nil));
	stt_varctx_set(ctx, "NEW-T", stt_create_builtin(ctx->stt, (stt_builtin_handler) stt_create_new_t));

	/* register error constants */
#define REG_ERR(CODE) \
	stt_regerrorcode(ctx->stt, #CODE); \
	stt_varctx_set(ctx, #CODE, stt_create_int(ctx->stt, STT_ ## CODE));

	REG_ERR(SE_UNDEF);
	REG_ERR(SE_WTF);
	REG_ERR(SE_PARM);
	REG_ERR(SE_WFUNC);
	REG_ERR(SE_TYPE);
	REG_ERR(SE_IPLIST);
	REG_ERR(SE_PNUM);
	REG_ERR(SE_IO);
	REG_ERR(SE_PARSE);
	REG_ERR(SE_NFND);
	REG_ERR(SE_LIMIT);
	REG_ERR(SE_MEM);
	REG_ERR(SE_MATH);
	REG_ERR(SE_RET);
	REG_ERR(SE_COMMA);
	REG_ERR(SE_COMMA_SPLICE);
	REG_ERR(SE_STOPREAD);

#include "builtin/basic_init.h"
#include "builtin/list_init.h"
#include "builtin/type_init.h"
#include "builtin/algebra_init.h"
#include "builtin/interp_init.h"
#include "builtin/str_init.h"
#include "builtin/io_init.h"
#include "builtin/array_init.h"
#include "builtin/dict_init.h"
#include "builtin/date_init.h"
#include "builtin/clos_init.h"
#include "builtin/os_init.h"
#include "builtin/popen_init.h"
#include "builtin/blob_init.h"
#include "builtin/sforms_init.h"
#include "builtin/math_init.h"


	stt_gc_prot_free(ctx->stt, mark, NULL);
}

void stt_register_stdio_as(stt_varctx_t *ctx, FILE *in, FILE *out, FILE *err)
{
	stt_obj_t *o;
	stt_s_mark mark;

	mark = stt_gc_prot_mark(ctx->stt);

	if (in != NULL) {
		o = stt_create_stream(ctx->stt, STT_STREAM_FP);
		o->d.stream.s.fp = in;
		stt_varctx_set(ctx, "*INPUT-STREAM*", o);
		stt_varctx_set(ctx, "*STDIN*", o);
	}

	if (out != NULL) {
		o = stt_create_stream(ctx->stt, STT_STREAM_FP);
		o->d.stream.s.fp = out;
		stt_varctx_set(ctx, "*OUTPUT-STREAM*", o);
		stt_varctx_set(ctx, "*STDOUT*", o);
	}

	if (err != NULL) {
		o = stt_create_stream(ctx->stt, STT_STREAM_FP);
		o->d.stream.s.fp = err;
		stt_varctx_set(ctx, "*STDERR*", o);
	}

	stt_gc_prot_free(ctx->stt, mark, NULL);
}

void stt_register_stdio(stt_varctx_t *ctx)
{
	stt_register_stdio_as(ctx, stdin, stdout, stderr);
}

