#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include "stt_config.h"
#include "memory.h"
#include "eval.h"
#include "compat.h"
#include "builtins.h"
#include "types.h"


#define GROWLIST_INIT(x) \
	x = (stt_obj_t **) stt_malloc(256 * sizeof(stt_obj_t *)); \
	x ## _size = 256; \
	x ## _tail = x; \
	x ## _cnt = 0;

#define GROWLIST_PAD(x) \
	if(x ## _cnt == x ## _size) { \
		x ## _size = x ## _size * 2; \
		((x) = (stt_obj_t **) stt_realloc(x, x ## _size * sizeof(stt_obj_t *))); \
		x ## _tail = x + x ## _cnt; \
	}

#define GROWLIST_ADD(x, o) \
	GROWLIST_PAD(x) \
	*(x ## _tail ++) = (o); \
	x ## _cnt ++;


#define GROWLIST_FREE(x) \
	free(x);

/* wrapper for calloc with error handling */
STT_INLINE void *stt_calloc(size_t nmemb, size_t size) {
	void *ptr = calloc(nmemb, size);
	if(!ptr) {
		perror("xcalloc");
		exit(STT_EXIT_MEMORY_FAIL);
	}
	return ptr;
}

/* wrapper for malloc with error handling */
STT_INLINE void *stt_malloc(size_t size) {
	void *ptr = malloc(size);
	if(!ptr) {
		perror("xmalloc");
		exit(STT_EXIT_MEMORY_FAIL);
	}
	return ptr;
}

/* wrapper for realloc with error handling */
STT_INLINE void *stt_realloc(void *ptr, size_t size) {
	ptr = realloc(ptr, size);
	if(!ptr) {
		perror("xrealloc");
		exit(STT_EXIT_MEMORY_FAIL);
	}
	return ptr;
}

/* add object to recycle list (and set its serial to 0) */
static void obj_recycle(stt_ctx_t *stt, stt_obj_t *o)
{
	o->serial = 0;
	GROWLIST_ADD(stt->recycle, o);
}

/* try to take an element of the recycle list */
stt_obj_t *stt_obj_respawn(stt_ctx_t *stt)
{
	if(stt->recycle_tail > stt->recycle) {
		stt->recycle_tail--;
		stt->recycle_cnt--;

		memset(*stt->recycle_tail, 0, sizeof(stt_obj_t));
		return *stt->recycle_tail;
	}

	return NULL;
}

/************************************** GARBAGE COLLECTOR ***/

/* list of protected roots */
static stt_protroot_t *gc_roots = NULL;

/* allocate stt_protroot_t structure */
static stt_protroot_t *gc_alloc_protroot(stt_varctx_t *ctx)
{
	stt_protroot_t *p;
	p = (stt_protroot_t *) stt_malloc(sizeof(stt_protroot_t));
	p->next = NULL;
	p->root = ctx;
	return p;
}

/* add an object to the root set */
void stt_gc_protect_obj(stt_obj_t *obj)
{
	stt_gc_protect_root((stt_varctx_t *) ((stt_intptr_t)obj | 1));
}

/* remove an object from the root set */
void stt_gc_unprotect_obj(stt_obj_t *obj)
{
	stt_gc_unprotect_root((stt_varctx_t *) ((stt_intptr_t)obj | 1));
}

/* protect a stt_varctx_t root (scan for variables in it) */
void stt_gc_protect_root(stt_varctx_t *ctx)
{
	stt_protroot_t *p;

	if(!gc_roots) {
		gc_roots = gc_alloc_protroot(ctx);
	} else {
		p = gc_alloc_protroot(ctx);
		p->next = gc_roots;
		gc_roots = p;
	}
}
/* unprotect a stt_varctx_t root (no longer scan for variables in it) */
void stt_gc_unprotect_root(stt_varctx_t *ctx)
{
	stt_protroot_t *p, *q;
	stt_protroot_t head;
	head.root = NULL;
	head.next = gc_roots;

	for(p = &head; p->next != NULL; ) {
		if(p->next->root == ctx) {
			q = p->next;
			p->next = p->next->next;
			free(q);
			break;
		}
		p = p->next;
	}
	gc_roots = head.next;
}

/* get current mark in protection stack */
stt_s_mark stt_gc_prot_mark(stt_ctx_t *stt) {
	return stt->gc_prot_cnt;
}

/* free up to mark in protection stack, except for object ret */
void stt_gc_prot_free(stt_ctx_t *stt, stt_s_mark mark, stt_obj_t *ret) {
	if(ret) {
		GROWLIST_PAD(stt->gc_prot)
		*(stt->gc_prot + mark) = ret;
		stt->gc_prot_cnt = mark + 1;
		stt->gc_prot_tail = stt->gc_prot + stt->gc_prot_cnt;
	} else {
		stt->gc_prot_cnt = mark;
		stt->gc_prot_tail = stt->gc_prot + stt->gc_prot_cnt;
	}
}

/* add object to protection list */
STT_INLINE void stt_gc_protlist_add(stt_ctx_t *stt, stt_obj_t *obj)
{
	GROWLIST_ADD(stt->gc_prot, obj);
}

/* record allocated object for GC */
void stt_gc_record(stt_ctx_t *stt, stt_obj_t *obj)
{
	GROWLIST_ADD(stt->gc_list, obj);
	stt->gc_allocd++;

	stt_gc_protlist_add(stt, obj);
}

/* remove last allocated object from GC records (in case you want to free it yourself) */
void stt_gc_unrecord(stt_ctx_t *stt)
{
	stt->gc_list_cnt--;
	stt->gc_list_tail--;
}

/* mark misc. pointer for free() */
void stt_mark_miscfree(stt_ctx_t *stt, void *ptr)
{
	GROWLIST_ADD(stt->miscfree, (stt_obj_t *) ptr);
}

/* misc pointers in miscfree list */
void stt_free_miscfree(stt_ctx_t *stt)
{
	stt_obj_t **n;
	for(n = stt->miscfree; n != stt->miscfree_tail; n++)
		free(*n);
	stt->miscfree_tail = stt->miscfree;
	stt->miscfree_cnt = 0;
}

/* checks whether the link should be followed
 * (not if the target object has the link in its stt_weakptr_t list */
STT_INLINE void stt_gc_follow(stt_ctx_t *stt, stt_obj_t **obj)
{
	stt_weakptr_t *wp;
	if(!*obj) return;
	for(wp = (*obj)->weak; wp; wp=wp->next) {
/*		fprintf(stderr, "follow? *obj=%p obj=%p ptr=%p *ptr=%p\n", *obj, obj, wp->ptr, *wp->ptr);*/
		if(wp->ptr == obj) {
/*			fprintf(stderr, "  no.\n");*/
			return;
		}
 	}

	stt_gc_mark_obj(stt, *obj);
}

/* mark variables reachable from object */
STT_INLINE void stt_gc_mark_obj(stt_ctx_t *stt, stt_obj_t *obj)
{
	if(!obj) return;
	if(obj->mark == stt->gc_mark) return; /* looked here already */
	obj->mark = stt->gc_mark;
	if(stt->gc_follow_dispatch[obj->type])
		stt->gc_follow_dispatch[obj->type](stt, obj);
}

/* GC follow: STT_ST_CONS */
static void gc_follow_cons(stt_ctx_t *stt, stt_obj_t *obj)
{
	stt_gc_follow(stt, &obj->d.cons.car);
	stt_gc_follow(stt, &obj->d.cons.cdr);
}

/* GC follow: STT_ST_ANNOT */
static void gc_follow_annot(stt_ctx_t *stt, stt_obj_t *obj)
{
	stt_gc_follow(stt, &obj->d.annot.val);
	stt_gc_follow(stt, &obj->d.annot.annot);
}

/* GC follow: STT_ST_VARCTX */
static void gc_follow_varctx(stt_ctx_t *stt, stt_obj_t *obj)
{
	stt_gc_mark(obj->d.varctx.ctx);
}

/* GC follow: STT_ST_FUNC */
static void gc_follow_func(stt_ctx_t *stt, stt_obj_t *obj)
{
	stt_gc_mark_obj(stt, obj->d.func.parm);
	stt_gc_mark_obj(stt, obj->d.func.exp);
	stt_gc_mark_obj(stt, obj->d.func.docstr);
	stt_gc_mark_obj(stt, obj->d.func.closure);
}

/* GC follow: STT_ST_DICT */
static void gc_follow_dict(stt_ctx_t *stt, stt_obj_t *obj)
{
	unsigned long i;
	for(i=0; i<obj->d.dict.size; i++) {
		if(obj->d.dict.data[i].name != NULL) {
			stt_gc_mark_obj(stt, obj->d.dict.data[i].name);
			stt_gc_follow(stt, &obj->d.dict.data[i].value);
		}
	}
}

/* GC follow: STT_ST_STREAM */
static void gc_follow_stream(stt_ctx_t *stt, stt_obj_t *obj)
{
	switch(obj->d.stream.type) {
		case STT_STREAM_FUNC:
			stt_gc_mark_obj(stt, obj->d.stream.s.func);
			break;
		default:
			;
	}
}

/* GC follow: STT_ST_ERROR */
static void gc_follow_error(stt_ctx_t *stt, stt_obj_t *obj)
{
	if(obj->d.error.obj) {
		stt_gc_mark_obj(stt, obj->d.error.obj);
	}
	if(obj->d.error.trace) {
		stt_gc_mark_obj(stt, obj->d.error.trace);
	}
}

/* GC follow: STT_ST_ARRAY */
static void gc_follow_array(stt_ctx_t *stt, stt_obj_t *obj)
{
	int i;
	long ds;
	stt_obj_t **a;

	ds = 1; /* compute array size */
	for(i=0; i<obj->d.array.dim; i++)
		ds *= obj->d.array.size[i];
	a = obj->d.array.data;
	for(i=0;i<ds;i++) {
		stt_gc_follow(stt, a);
		a++;
	}
}

/* GC follow: STT_ST_CLASS */
static void gc_follow_class(stt_ctx_t *stt, stt_obj_t *obj)
{
	size_t i;

	stt_gc_mark_obj(stt, obj->d.clos_class->parents);
	stt_gc_mark_obj(stt, obj->d.clos_class->children);
	stt_gc_mark_obj(stt, obj->d.clos_class->slots);
	stt_gc_mark_obj(stt, obj->d.clos_class->prec);
	stt_gc_mark_obj(stt, obj->d.clos_class->name);
	stt_gc_mark_obj(stt, obj->d.clos_class->ancestry);

	for(i=0; i<obj->d.clos_class->allslotsnum; i++) {
		stt_gc_mark_obj(stt, obj->d.clos_class->allslots[i].stor);
		stt_gc_mark_obj(stt, obj->d.clos_class->allslots[i].slot);
	}

	for(i=0; i<obj->d.clos_class->data_size; i++)
		stt_gc_follow(stt, &obj->d.clos_class->data[i]);

	stt_gc_mark(obj->d.clos_class->initarg);
	stt_gc_mark(obj->d.clos_class->slotname);
}

/* GC follow: STT_ST_SLOT */
static void gc_follow_slot(stt_ctx_t *stt, stt_obj_t *obj)
{
	stt_gc_mark_obj(stt, obj->d.clos_slot.initform);
	stt_gc_mark_obj(stt, obj->d.clos_slot.name);
	stt_gc_mark_obj(stt, obj->d.clos_slot.accessor);
	stt_gc_mark_obj(stt, obj->d.clos_slot.writer);
	stt_gc_mark_obj(stt, obj->d.clos_slot.reader);
	stt_gc_mark_obj(stt, obj->d.clos_slot.initarg);
}

/* GC follow: STT_ST_OBJECT */
static void gc_follow_object(stt_ctx_t *stt, stt_obj_t *obj)
{
	size_t i;
	stt_gc_mark_obj(stt, obj->d.clos_object.type);
	for(i=0; i<obj->d.clos_object.data_size; i++)
		stt_gc_follow(stt, &obj->d.clos_object.data[i]);
}

/* GC follow: STT_ST_GENERIC */
static void gc_follow_generic(stt_ctx_t *stt, stt_obj_t *obj)
{
	int i;
	stt_gc_mark_obj(stt, obj->d.clos_generic.m_primary);
	stt_gc_mark_obj(stt, obj->d.clos_generic.m_before);
	stt_gc_mark_obj(stt, obj->d.clos_generic.m_after);
	stt_gc_mark_obj(stt, obj->d.clos_generic.m_around);

	for(i=0; i<obj->d.clos_generic.dispatch_size; i++) {
		stt_dispatch_signature_t *sgnt;

		stt_gc_mark_obj(stt, obj->d.clos_generic.dispatch[i].method);

		for(sgnt = obj->d.clos_generic.dispatch[i].signature; sgnt && sgnt->type != -1 ; sgnt++)
			stt_gc_mark_obj(stt, sgnt->value);

	}
}

/* GC follow: STT_ST_METHOD */
static void gc_follow_method(stt_ctx_t *stt, stt_obj_t *obj)
{
	int i;

	for(i=0; i<obj->d.clos_method.argn; i++) {
		stt_gc_mark_obj(stt, obj->d.clos_method.args[i].name);
		stt_gc_mark_obj(stt, obj->d.clos_method.args[i].value);
	}

	stt_gc_mark_obj(stt, obj->d.clos_method.miscargs);
	stt_gc_mark_obj(stt, obj->d.clos_method.code);
	stt_gc_mark_obj(stt, obj->d.clos_method.next);
}

/* GC follow: STT_ST_DISPATCH */
static void gc_follow_st_dispatch(stt_ctx_t *stt, stt_obj_t *obj)
{
	int i;
	stt_gc_mark_obj(stt, obj->d.clos_dispatch.generic);
	stt_gc_mark_obj(stt, obj->d.clos_dispatch.method);
	i=0;
	while(obj->d.clos_dispatch.argeval[i]) {
		stt_gc_mark_obj(stt, obj->d.clos_dispatch.argeval[i++]);
	}
}

/* mark variables in context */
void stt_gc_mark(stt_varctx_t *ctx)
{
	long i;
	stt_var_t *var;

	stt_gc_mark_obj(ctx->stt, ctx->ref);
	if(ctx->parent) stt_gc_mark(ctx->parent);

	if(ctx->mark == ctx->stt->gc_mark)
		return; /* why hello, we looked here already */
	ctx->mark = ctx->stt->gc_mark;

	for(i=0; i<ctx->ns; i++) {
		for(var = ctx->hash[i]; var; var = var->next) {
			if(var->value) {
				stt_gc_follow(ctx->stt, &var->value);
			}
		}
	}
}

/* free all data associated with object */
STT_INLINE void stt_gc_harvest(stt_ctx_t *stt, stt_obj_t *obj)
{
	stt_weakptr_t *ptr, *n;

	/* set all weak pointers to nil, but only if they still point here */
	for(ptr = obj->weak; ptr;) {
		if(ptr->host->serial == ptr->serial && /* host object still exists */
			*ptr->ptr == obj) {

			*ptr->ptr = stt_create_nil(stt);
		}

		n = ptr->next;
		free(ptr);
		ptr = n;
	}

	obj->weak = NULL;

	stt_freedata(obj);
	obj_recycle(stt, obj);
}

/* do garbage collection. returns number of objects freed */
int stt_gc_sweep(stt_ctx_t *stt, stt_varctx_t *ctx)
{
	stt_obj_t **a;
	stt_protroot_t *p;
	int sw = 0;

	if(stt->gc_disabled) return 0;

	/* mark all reachable objects */
	stt->gc_mark++;
	if (ctx) {
		assert(stt == ctx->stt);
		stt_gc_mark(ctx);
	}

	for(p = gc_roots; p; p=p->next)
		if((stt_intptr_t) (p->root) & 1)
			stt_gc_mark_obj(stt, (stt_obj_t *) ((stt_intptr_t) (p->root) ^ 1));
		else
			stt_gc_mark(p->root);

	for(a = stt->gc_prot; a != stt->gc_prot_tail; a++) {
		stt_gc_mark_obj(stt, *a);
	}

	/* well, all classes referenced in type db should be
	 * registered in stt->class_hash anyway */
/*	for(int i = 0; i < stt->next_typeid; i++)
			stt_gc_mark_obj(stt, stt->typedb[i].cl);*/

	/* sweep generation 1 */
	for(a = stt->gc_list; a != stt->gc_list_tail; ) {
		if((*a)->mark < stt->gc_mark) {

			sw++;
			stt_gc_harvest(stt, *a);
			stt->gc_list_tail --;
			*a = *stt->gc_list_tail;
			stt->gc_list_cnt --;
		} else {

			/* move object to generation 2 */

			GROWLIST_ADD(stt->gc_gen2, *a);

			stt->gc_list_tail --;
			*a = *stt->gc_list_tail;
			stt->gc_list_cnt --;
		}
	}

	stt->gc_gen2_count --;
	if(stt->gc_gen2_count == 0) {
		stt->gc_gen2_count = GC_GEN2_PERIOD;
		stt->gc_gen3_count--;

		for(a = stt->gc_gen2; a != stt->gc_gen2_tail; ) {
			if((*a)->mark < stt->gc_mark) {

				sw++;
				stt_gc_harvest(stt, *a);
				stt->gc_gen2_tail --;
				*a = *stt->gc_gen2_tail;
				stt->gc_gen2_cnt --;
			} else {

				/* move object to generation 3 */

				GROWLIST_ADD(stt->gc_gen3, *a);

				stt->gc_gen2_tail --;
				*a = *stt->gc_gen2_tail;
				stt->gc_gen2_cnt --;
			}
		}

		if(stt->gc_gen3_count == 0) {
			stt->gc_gen3_count = GC_GEN3_PERIOD;

			for(a = stt->gc_gen3; a != stt->gc_gen3_tail; ) {
				if((*a)->mark < stt->gc_mark) {

					sw ++;
					stt_gc_harvest(stt, *a);
					stt->gc_gen3_tail --;
					*a = *stt->gc_gen3_tail;
					stt->gc_gen3_cnt--;
				} else a++;
			}
		}
	}

	stt->gc_live += (stt->gc_allocd - sw);

	if(stt->gc_verbose) {
		fprintf(stderr, "GC: %d freed, %ld estimated live (%d %d %d), pool=%d\n", sw, stt->gc_live, stt->gc_list_cnt, stt->gc_gen2_cnt, stt->gc_gen3_cnt, stt->recycle_cnt);
	}

	stt->gc_allocd = 0;

	return sw;
}

/**************************************** VARIABLE CONTEXTS ***/

/* create variable context */
stt_varctx_t *stt_varctx_create(stt_ctx_t *stt, stt_varctx_t *parent, const long size)
{
	stt_varctx_t *vc;

	assert(!((size - 1) & size));

	vc = (stt_varctx_t *) stt_malloc(sizeof(stt_varctx_t));
	vc->stt = stt;
	vc->parent = parent;
	vc->ns = size;
	vc->nsm = size - 1;
	vc->id = stt->varctx_id++;
	vc->weak = 0;
	vc->ref = NULL;
	vc->mark = 0;

	if(size > 0) {
		vc->defer = 0;
		vc->hash = (stt_var_t **) stt_calloc(size, sizeof(stt_var_t *));
	} else {
		vc->defer = 1;
	}

	return vc;
}

static stt_var_t *alloc_s_var(void)
{
	stt_var_t *s;
	s = (stt_var_t *) stt_calloc(1, sizeof(stt_var_t));
	return s;
}

/* store a variable inside a ctx */
void stt_varctx_store_s(stt_varctx_t *ctx, unsigned long sdbmv, unsigned long id, stt_obj_t *val)
{
	stt_var_t *sv;
	stt_var_t **sf;

	sv = alloc_s_var();
	sv->name = NULL;
	sv->sdbm = sdbmv;
	sv->id = id;
	sv->value = val;

	sf = &(ctx->hash[sdbmv & ctx->nsm]);
	while(*sf) sf = &((*sf)->next);
	*sf = sv;
}

/* store a variable inside a ctx */
void stt_varctx_store(stt_varctx_t *ctx, char *name, stt_obj_t *val)
{
	unsigned long sdbmv = stt_sdbm(name);
	stt_varctx_store_s(ctx, sdbmv, stt_get_symbolid(ctx->stt, name, sdbmv), val);
}

/* store a variable inside a ctx - with given sdbm value */
void stt_varctx_store_sdbm(stt_varctx_t *ctx, char *name, stt_obj_t *val, unsigned long sdbmv)
{
	stt_varctx_store_s(ctx, sdbmv, stt_get_symbolid(ctx->stt, name, sdbmv), val);
}

/* empty a stt_varctx_t - erase all lists */
static void varctx_empty(stt_varctx_t *ctx)
{
	long i;
	stt_var_t *v, *f;

	for(i=0; i<ctx->ns; i++) {
		for(v = ctx->hash[i]; v;) {
/*			stt_weak_unref(&v->value);*/
			f = v;
			v = v->next;
			free(f->name);
			free(f);
		}
		ctx->hash[i] = NULL;
	}
}

/* get symbol name by symbol id */
char * stt_get_symbolname(stt_ctx_t *stt, unsigned long sdbmv, unsigned long id)
{
	stt_var_t *sv;

	sv = stt->symbol_name_hash->hash[sdbmv & stt->symbol_name_hash->nsm];
	while(sv) {
		if(sv->id == id)
			return sv->name;
		sv = sv->next;
	}

	return NULL;
}

/* get hash stt_var_t of a symbol */
stt_var_t * stt_get_symbolsv(stt_ctx_t *stt, const char *name, unsigned long sdbmv)
{
	stt_var_t *sv, **sf;
	char *c;

	if(!stt->symbol_name_hash) {
		stt->symbol_name_hash = stt_varctx_create(stt, NULL, SYMBOL_NAME_HASH_SIZE);
		stt_gc_protect_root(stt->symbol_name_hash);
	}

	sv = stt->symbol_name_hash->hash[sdbmv & stt->symbol_name_hash->nsm];
	while(sv) {
		if(!stt_strcasecmp(name, sv->name))
			return sv;
		sv = sv->next;
	}

	sv = alloc_s_var();
	sv->name = stt_strdup(name);

	for(c = sv->name; *c; c++) *c = toupper(*c);
	sv->sdbm = sdbmv;
	sv->id = stt->symbol_id ++;

	sf = &(stt->symbol_name_hash->hash[sdbmv & stt->symbol_name_hash->nsm]);
	while(*sf) sf = &((*sf)->next);
	*sf = sv;

	return sv;
}

/* get unique symbol id of a symbol */
unsigned long stt_get_symbolid(stt_ctx_t *stt, const char *name, unsigned long sdbmv)
{
	stt_var_t *sv = stt_get_symbolsv(stt, name, sdbmv);
	return sv->id;
}

/* utility function: get symbol id by string */
unsigned long stt_symbolid(stt_ctx_t *stt, char *name)
{
	return stt_get_symbolid(stt, name, stt_sdbm(name));
}

/* get value of a variable from a ctx (local only)
 * returns NULL if variable doesn't exist */
stt_var_t *stt_varctx_retr_s(stt_varctx_t *ctx, unsigned long sdbmv, unsigned long id)
{
	stt_var_t *sv;
	sv = ctx->hash[sdbmv & ctx->nsm];
	while(sv) {
		if(id == sv->id)
			return sv;
		sv = sv->next;
	}

	return NULL;
}

/* get value of a variable from a ctx (local only)
 * returns NULL if variable doesn't exist */
stt_var_t *stt_varctx_retr(stt_varctx_t *ctx, char *name)
{
	unsigned long sdbmv = stt_sdbmc(name);
	return stt_varctx_retr_s(ctx, sdbmv, stt_get_symbolid(ctx->stt, name, sdbmv));
}

/* get value of a variable from a ctx (local only)
 * returns NULL if variable doesn't exist */
stt_var_t *stt_varctx_retr_sdbm(stt_varctx_t *ctx, char *name, unsigned long sdbmv)
{
	return stt_varctx_retr_s(ctx, sdbmv, stt_get_symbolid(ctx->stt, name, sdbmv));
}

/* get value of a variable from a ctx (with parent fallback)
 * returns NULL if variable doesn't exist */
stt_var_t *stt_varctx_get_s(stt_varctx_t *ctx, unsigned long sdbmv, unsigned long id)
{
	stt_var_t *v;
	v = stt_varctx_retr_s(ctx, sdbmv, id);
	if(!v && ctx->parent) v = stt_varctx_get_s(ctx->parent, sdbmv, id);
	return v;
}

/* get value of a variable from a ctx (with parent fallback)
 * returns NULL if variable doesn't exist */
stt_var_t *stt_varctx_get_sdbm(stt_varctx_t *ctx, char *name, unsigned long sdbmv)
{
	return stt_varctx_get_s(ctx, sdbmv, stt_get_symbolid(ctx->stt, name, sdbmv));
}

/* get value of a variable from a ctx (with parent fallback)
 * returns NULL if variable doesn't exist */
stt_var_t *stt_varctx_get(stt_varctx_t *ctx, char *name)
{
	return stt_varctx_get_sdbm(ctx, name, stt_sdbmc(name));
}

/* set/store a variable inside a ctx
 * (updates if the var already exists) */
void stt_varctx_set_s(stt_varctx_t *ctx, unsigned long sdbmv, unsigned long id, stt_obj_t *val)
{
	stt_var_t *v;
	if((v = stt_varctx_retr_s(ctx, sdbmv, id))) {
		v->value = val;
		return;
	}

	if(ctx->weak && ctx->parent) stt_varctx_set_s(ctx->parent, sdbmv, id, val);
	else stt_varctx_store_s(ctx, sdbmv, id, val);
}

/* set/store a variable inside a ctx
 * (updates if the var already exists) */
void stt_varctx_set_sdbm(stt_varctx_t *ctx, char *name, stt_obj_t *val, unsigned long sdbmv)
{
	stt_var_t *v;
	if((v = stt_varctx_retr_sdbm(ctx, name, sdbmv))) {
		v->value = val;
		return;
	}

	if(ctx->weak && ctx->parent) stt_varctx_set_sdbm(ctx->parent, name, val, sdbmv);
	else stt_varctx_store_sdbm(ctx, name, val, sdbmv);
}

/* set/store a variable inside a ctx
 * (updates if the var already exists) */
void stt_varctx_set(stt_varctx_t *ctx, const char *name, stt_obj_t *val)
{
	stt_var_t *v;
	unsigned long sdbmv = stt_sdbmc(name);
	unsigned long id = stt_get_symbolid(ctx->stt, name, sdbmv);

	if((v = stt_varctx_retr_s(ctx, sdbmv, id))) {
		v->value = val;
		return;
	}

	if(val->type == STT_ST_BUILTIN) {
		if(!val->d.builtin.symbol) {
			val->d.builtin.symbol = stt_get_symbolname(ctx->stt, sdbmv, id);
			val->d.builtin.symbolid = id;
		}
	}

	stt_varctx_store_s(ctx, sdbmv, id, val);
}

/* compute hash value for a string */
unsigned long stt_sdbm(const char *str)
{
	unsigned long hash = 0;
	int c;

	while((c = *str++))
		hash = c + (hash << 6) + (hash << 16) - hash;

	return hash;
}

/* compute hash value for a string, ignoring case */
unsigned long stt_sdbmc(const char *str)
{
	unsigned long hash = 0;
	int c;

	while((c = toupper(*str++)))
		hash = c + (hash << 6) + (hash << 16) - hash;

	return hash;
}

/* free stt_varctx_t */
void stt_varctx_free(stt_varctx_t *ctx)
{
	stt_gc_unprotect_root(ctx);

	if(!ctx->defer) {
		varctx_empty(ctx);
		free(ctx->hash);
	}

	free(ctx);
}

/* dump contents of stt_varctx_t to stderr */
stt_obj_t * stt_varctx_dump(stt_varctx_t *ctx)
{
	long i;
	stt_var_t *var;
	stt_obj_t *o;

	o = stt_stream_printf(ctx, "---------- %p\n", ctx);
	if(o->type == STT_ST_ERROR) return o;

	while(ctx->parent) {
		for(i=0; i<ctx->ns; i++) {
			for(var = ctx->hash[i]; var; var = var->next) {
				o = stt_stream_printf(ctx, "%p %s = ", var->value, stt_get_symbolname(ctx->stt, var->sdbm, var->id));
				if(o->type == STT_ST_ERROR) return o;
				o = stt_printobj(ctx, var->value);
				if(o->type == STT_ST_ERROR) return o;
				o = stt_stream_stdout_write(ctx, '\n');
				if(o->type == STT_ST_ERROR) return o;
			}
		}

		o = stt_stream_printf(ctx, "========== %p\n", ctx->parent);
		if(o->type == STT_ST_ERROR) return o;
		ctx = ctx->parent;
	}
	o = stt_stream_printf(ctx, "---------------------------\n");
	if(o->type == STT_ST_ERROR) return o;
	return stt_create_nil(ctx->stt);
}

/****************************************** INIT ***/

/* initialize Stutter */
void stt_init(stt_ctx_t *stt)
{
	int i;
	const char *userref;

	if (sizeof(void *) > sizeof(stt_intptr_t)) {
		fprintf(stderr, "estutter build error: stt_intptr_t too small\n");
		abort();
	}

	stt->load_path = ".";

	stt->trace_depth = 0;
	stt->file = "stdin";
	stt->lastcall_file = NULL;
	stt->lastcall_line = 0;
	stt->parsed_line = 0;
	stt->recursion_limit = 1000;
	stt->next_serial = 1;
	stt->default_dict_size = DEFAULT_DICT_SIZE;
	stt->dict_threshold = DICT_THRESHOLD;

	stt->gc_allocd = 0;
	stt->gc_live = 0;
	stt->gc_gen2_count = GC_GEN2_PERIOD;
	stt->gc_gen3_count = GC_GEN3_PERIOD;
	stt->gc_mark = 1;
	stt->gc_allocd = 0;
	stt->gc_alloc_limit = STT_GC_DEFAULT_ALLOC_LIMIT;
	stt->gc_verbose = 0;
	stt->gc_disabled = 0;

	stt->varctx_id = 1;
	stt->symbol_id = 1;
	stt->macro_annot = 1;

	stt->errornames = NULL;
	stt->errorcount = 0;

	stt->symbol_name_hash = NULL;
	stt->rest_sdbm = stt_sdbm("_REST");
	stt->rest_id = stt_get_symbolid(stt, "_REST", stt->rest_sdbm);

	stt->dispatch_info_sdbm = stt_sdbm("_DISPATCH-INFO");
	stt->dispatch_info_id = stt_get_symbolid(stt, "_DISPATCH-INFO", stt->dispatch_info_sdbm);

	stt->lambda_sdbm = stt_sdbm("LAMBDA");
	stt->lambda_id = stt_get_symbolid(stt, "LAMBDA", stt->lambda_sdbm);

	stt->recurse_sdbm = stt_sdbm("<^");
	stt->recurse_id = stt_get_symbolid(stt, "<^", stt->recurse_sdbm);

	GROWLIST_INIT(stt->gc_prot);
	GROWLIST_INIT(stt->gc_list);
	GROWLIST_INIT(stt->gc_gen2);
	GROWLIST_INIT(stt->gc_gen3);
	GROWLIST_INIT(stt->miscfree);
	GROWLIST_INIT(stt->recycle);

	stt_objcache_init(stt);

	stt->class_hash = stt_varctx_create(stt, NULL, 64);
	stt_gc_protect_root(stt->class_hash);

	stt->next_typeid = 0;
	stt->typedb = NULL;
	for(i=0; i<ST_TYPE_LAST_MARK; i++)
		stt_gettypeid(stt);
	stt->class_precedence = NULL;

	/* set up garbage collector */
	for(i=0; i<ST_TYPE_LAST_MARK; i++)
		stt->gc_follow_dispatch[i] = NULL;
	stt->gc_follow_dispatch[STT_ST_CONS] = gc_follow_cons;
	stt->gc_follow_dispatch[STT_ST_ANNOT] = gc_follow_annot;
	stt->gc_follow_dispatch[STT_ST_VARCTX] = gc_follow_varctx;
	stt->gc_follow_dispatch[STT_ST_FUNC] = gc_follow_func;
	stt->gc_follow_dispatch[STT_ST_DICT] = gc_follow_dict;
	stt->gc_follow_dispatch[STT_ST_STREAM] = gc_follow_stream;
	stt->gc_follow_dispatch[STT_ST_ERROR] = gc_follow_error;
	stt->gc_follow_dispatch[STT_ST_ARRAY] = gc_follow_array;
	stt->gc_follow_dispatch[STT_ST_CLASS] = gc_follow_class;
	stt->gc_follow_dispatch[STT_ST_SLOT] = gc_follow_slot;
	stt->gc_follow_dispatch[STT_ST_OBJECT] = gc_follow_object;
	stt->gc_follow_dispatch[STT_ST_GENERIC] = gc_follow_generic;
	stt->gc_follow_dispatch[STT_ST_METHOD] = gc_follow_method;
	stt->gc_follow_dispatch[STT_ST_DISPATCH] = gc_follow_st_dispatch;

	/* set up refpath */
	stt->refpath = stt_strbuild_new();
	userref = getenv("STT_REFPATH");
	if (userref == NULL)
		userref = getenv("STUTTER_HOME"); /* compatibility with stutter */
	if(userref) {
		stt_strbuild_addstr(&stt->refpath, userref, strlen(userref));
	}
	else {
#ifdef STT_REFPATH
		stt_strbuild_addstr(&stt->refpath, STT_REFPATH, strlen(STT_REFPATH));
#else
		char cwd[1024];
		if(getcwd(cwd, sizeof(cwd)) != NULL)
			stt_strbuild_addstr(&stt->refpath, cwd, strlen(cwd));
		else
			stt_strbuild_addstr(&stt->refpath, ".", 1);
#endif
	}
}

/* shut down Stutter */
void stt_uninit(stt_ctx_t *stt)
{
	stt_obj_t **a;

	stt_gc_unprotect_root(stt->class_hash);
	stt_varctx_free(stt->symbol_name_hash);
	stt_gc_prot_free(stt, 0, NULL);
	stt->gc_gen2_count = 1;
	stt->gc_gen3_count = 1;
	stt_gc_sweep(stt, NULL);
	stt_varctx_free(stt->class_hash);
	stt_free_miscfree(stt);
	free(stt->typedb);
	free(stt->refpath.str);
	stt->refpath.len = stt->refpath.buf = 0;

	for(a = stt->recycle; a != stt->recycle_tail; a++) {
		free(*a);
	}

	GROWLIST_FREE(stt->gc_prot);
	GROWLIST_FREE(stt->gc_list);
	GROWLIST_FREE(stt->gc_gen2);
	GROWLIST_FREE(stt->gc_gen3);
	GROWLIST_FREE(stt->miscfree);
	GROWLIST_FREE(stt->recycle);
	free(stt->errornames);
}

/********************************** newhash */

typedef  unsigned long int  u4;   /* unsigned 4-byte type */
typedef  unsigned     char  u1;   /* unsigned 1-byte type */

/* The mixing step */
#define mix(a,b,c) \
{ \
	a=a-b;  a=a-c;  a=a^(c>>13); \
	b=b-c;  b=b-a;  b=b^(a<<8);  \
	c=c-a;  c=c-b;  c=c^(b>>13); \
	a=a-b;  a=a-c;  a=a^(c>>12); \
	b=b-c;  b=b-a;  b=b^(a<<16); \
	c=c-a;  c=c-b;  c=c^(b>>5);  \
	a=a-b;  a=a-c;  a=a^(c>>3);  \
	b=b-c;  b=b-a;  b=b^(a<<10); \
	c=c-a;  c=c-b;  c=c^(b>>15); \
}

/* The whole new hash function */
u4 stt_hash(k, length, initval)
	register u1 *k;        /* the key */
	u4           length;   /* the length of the key in bytes */
	u4           initval;  /* the previous hash, or an arbitrary value */
{
	register u4 a,b,c;  /* the internal state */
	u4          len;    /* how many key bytes still need mixing */

	/* Set up the internal state */
	len = length;
	a = b = 0x9e3779b9;  /* the golden ratio; an arbitrary value */
	c = initval;         /* variable initialization of internal state */

	/*---------------------------------------- handle most of the key */
	while (len >= 12)
	{
		a=a+(k[0]+((u4)k[1]<<8)+((u4)k[2]<<16) +((u4)k[3]<<24));
		b=b+(k[4]+((u4)k[5]<<8)+((u4)k[6]<<16) +((u4)k[7]<<24));
		c=c+(k[8]+((u4)k[9]<<8)+((u4)k[10]<<16)+((u4)k[11]<<24));
		mix(a,b,c);
		k = k+12; len = len-12;
	}

	/*------------------------------------- handle the last 11 bytes */
	c = c+length;
	switch(len)              /* all the case statements fall through */
	{
		case 11: c=c+((u4)k[10]<<24);
		case 10: c=c+((u4)k[9]<<16);
		case 9 : c=c+((u4)k[8]<<8);
				 /* the first byte of c is reserved for the length */
		case 8 : b=b+((u4)k[7]<<24);
		case 7 : b=b+((u4)k[6]<<16);
		case 6 : b=b+((u4)k[5]<<8);
		case 5 : b=b+k[4];
		case 4 : a=a+((u4)k[3]<<24);
		case 3 : a=a+((u4)k[2]<<16);
		case 2 : a=a+((u4)k[1]<<8);
		case 1 : a=a+k[0];
				 /* case 0: nothing left to add */
	}
	mix(a,b,c);
	/*-------------------------------------------- report the result */
	return c;
}

